#!/usr/bin/env python
"""A Python bridge to remotely control the LAASWM windows manager.

This library use the standard Python logging mechanism.
You can retrieve laaswm log messages through the "laaswm" logger. See the end of
this file for an example showing how to display to the console the log messages.
"""
import logging
import socket
import re

class NullHandler(logging.Handler):
    """Defines a NullHandler for logging, in case laaswm is used in an application
    that doesn't use logging.
    """
    def emit(self, record):
        pass

laaswmlogger = logging.getLogger("laaswm")

h = NullHandler()
laaswmlogger.addHandler(h)

class LaasWMError(Exception):
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class WM():
    def __init__(self, host = "localhost", port = 5554):

        self._laaswm_server = None
        
        try:
            #create an INET, STREAMing socket
            self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            #now connect to the oro server
            self.s.connect((host, port))
            self._laaswm_server = self.s.makefile(mode='rw')
        except socket.error:
            self.s.close()
            raise LaasWMError('Unable to connect to the windows manager. Check it is running and ' + \
                                 'that you provided the right host and port.')

        # Stores windows geometry before going fullscreen for later
        # restoration
        self.windows = self.list()
        
        self.currentgamma = 1.0
        self.gamma(1.0)
    
    def __del__(self):
        self.close()

    def close(self):
        if self._laaswm_server:
            self._laaswm_server.close()
        if self.s:
            self.s.close()


    def rpc(self, req):
        laaswmlogger.debug("Send request: " + req)
        print("Send request: " + req)
        self._laaswm_server.write(req)
        self._laaswm_server.write("\n")
        self._laaswm_server.flush()
        res = []
        line = ""
        while True:
            line = self._laaswm_server.readline().strip()
            print("Got: " + line)
            if line == "ERROR":
                raise LaasWMError("The windows manager returned an error while trying to %s."%(req))
            if line == "OK":
                break
            res.append(line.split(req + " ")[1])
        return res

    def list(self):

        resolution = self.resolution()

        laaswmlogger.info("Listing remote windows...")
        p = re.compile("(\*| ) (0x[0-9a-f]+) (.+) \(\d+, \d+\) @ (\d+),(\d+) (\d+)x(\d+)")
        raw = self.rpc("LIST")
        res = {}
        for win in raw:
            m = p.match(win)

            name = m.group(3)
            hasfocus = True if "*" in m.group(1) else False
            x = int(m.group(4))
            y = int(m.group(5))
            width = int(m.group(6))
            height = int(m.group(7))

            laaswmlogger.info( "%s: %dx%d @(%d,%d)"%(name, width, height, x, y))

            res[m.group(2)] = { "name": name,
                                "width": width, 
                                "height": height, 
                                "x": x,
                                "y": y,
                                "hasfocus": hasfocus,
                                "isvisible": True,
                                "isfullscreen": True if [x,y] == resolution else False
                            }
        return res

    def setfocused(self, winid):
        if (self.focused() != None):
            self.windows[self.focused()]["hasfocus"] = False
            self.windows[winid]["hasfocus"] = True

    def focused(self):
        """ Returns the name of the window that currently own the focus.
        """
        for id, win in self.windows.items():
            if win["hasfocus"]:
                return id
        return None

    def gamma(self, gamma):
        if abs(self.currentgamma - gamma) > 0.1:
             self.currentgamma = gamma
             laaswmlogger.info("Setting gamma to %d..."%(gamma))
             try:
                 self.rpc("GAMMA " + str(gamma))
             except LaasWMError as e:
                 laaswmlogger.error(e.value)
                 return False
        return True

    def move(self, name, x, y):
        laaswmlogger.info("Moving remote window %s to (%d,%d)..."%(name,x,y))
        try:
            self.setfocused(name)
            self.rpc("MOVE %s %d %d"%(name,x,y))
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True
        laaswmlogger.info("Moving remote window %s to (%d,%d)..."%(name,x,y))
        try:
            self.setfocused(name)
            self.rpc("MOVE %s %d %d"%(name,x,y))
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True

    def geom(self, name, x, y, width, height):
        laaswmlogger.info("Changing geometry of remote window %s to %dx%d @(%d,%d)..."%(name,width,height,x,y))
        try:
            self.setfocused(name)
            self.rpc("GEOM %s %d %d %d %d"%(name,x,y, width, height))
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True

    def resolution(self):
        laaswmlogger.info("Querying for screen resolution...")
        res = self.rpc("SCREEN")

        return [int(x) for x in res[0].split()]

    def fullscreen(self, name):
        laaswmlogger.info("Bringing %s to fullscreen..."%(name))
        resolution = self.resolution()
        self.windows = self.list()

        self.geom(name, 0, 0, *resolution)

    def restore(self, name):
        laaswmlogger.info("Restoring %s..."%(name))

        self.geom(name,
                  self.windows[name]["x"],
                  self.windows[name]["y"],
                  self.windows[name]["width"],
                  self.windows[name]["height"])

    def fadein(self, name):
        laaswmlogger.info("Fading in window %s..."%(name))
        try:
            self.rpc("FADE %s 0"%(name))
            self.windows[name]["isvisible"] = True
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True

    def fadeout(self, name):
        laaswmlogger.info("Fading out window %s..."%(name))
        try:
            self.rpc("FADE %s 1"%(name))
            self.windows[name]["isvisible"] = False
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True

    def crossfade(self, name1, name2):
        laaswmlogger.info("Cross fading windows %s and %s..."%(name1, name2))
        try:
            self.setfocused(name2)
            self.rpc("FADE2 %s %s"%(name1, name2))
            self.setfocused(name1)
            self.windows[name1]["isvisible"] = True
            self.windows[name2]["isvisible"] = False
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True

    def kill(self, name):
        laaswmlogger.info("Killing remote window %s..."%(name))
        try:
            self.rpc("KILL %s"%(name))
        except LaasWMError as e:
            laaswmlogger.error(e.value)
            return False
        return True

    def execute(self, cmd):
        laaswmlogger.info("Executing cmd \"%s\" on remote host..."%(cmd))
        self.rpc("NEW %s"%(cmd))
        return True

if __name__ == '__main__':

    from random import randint

    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)

    # set a format which is simpler for console use
    formatter = logging.Formatter('%(asctime)-15s %(name)s: %(levelname)s - %(message)s')
    # tell the handler to use this format
    console.setFormatter(formatter)
    # add the handler to the root logger
    laaswmlogger.addHandler(console)
    
    HOST = 'localhost'    # server host
    PORT = 5554        # server port
    
    wm = WM(HOST, PORT)
    
   
    laaswmlogger.info("Testing LAASWM bridge now...")

    try:
       windows = wm.list()
       for w in windows.keys():
           wm.move(w, randint(10, 200), randint(10,200))

       laaswmlogger.info("done!")
        
        
    except LaasWMError as ose:
        print('Oups! An error occured!')
        print(ose)
    finally:
        wm.close()
